@@ -19,8 +19,9 @@
 #include "tcp.h"
 #include "util.h"
 #include "nids.h"
+#include "hash.h"
 
-#ifndef __linux__
+#if ! HAVE_TCP_STATES
 enum {
   TCP_ESTABLISHED = 1,
   TCP_SYN_SENT,
@@ -35,12 +36,7 @@ enum {
   TCP_CLOSING			/* now a valid state */
 };
 
-#define ICMP_DEST_UNREACH ICMP_UNREACH
-#define ICMP_PROT_UNREACH ICMP_UNREACH_PROTOCOL
-#define ICMP_PORT_UNREACH ICMP_UNREACH_PORT
-#define NR_ICMP_UNREACH   ICMP_MAXTYPE
-
-#endif				/* !linux */
+#endif
 
 #define FIN_SENT 120
 #define FIN_CONFIRMED 121
@@ -57,18 +53,34 @@ static struct tcp_stream **tcp_stream_table;
 static struct tcp_stream *streams_pool;
 static int tcp_num = 0;
 static int tcp_stream_table_size;
-static int hash_deg;
 static int max_stream;
 static struct tcp_stream *tcp_latest = 0, *tcp_oldest = 0;
 static struct tcp_stream *free_streams;
 static struct ip *ugly_iphdr;
 
+static void purge_queue(struct half_stream * h)
+{
+  struct skbuff *tmp, *p = h->list;
+
+  while (p) {
+    free(p->data);
+    tmp = p->next;
+    free(p);
+    p = tmp;
+  }
+  h->list = h->listtail = 0;
+  h->rmem_alloc = 0;
+}
+
 static void
 free_tcp(struct tcp_stream * a_tcp)
 {
   int hash_index = a_tcp->hash_index;
   struct lurker_node *i, *j;
-  
+ 
+  purge_queue(&a_tcp->server);
+  purge_queue(&a_tcp->client);
+   
   if (a_tcp->next_node)
     a_tcp->next_node->prev_node = a_tcp->prev_node;
   if (a_tcp->prev_node)
@@ -103,19 +115,8 @@ free_tcp(struct tcp_stream * a_tcp)
 static int
 mk_hash_index(struct tuple4 addr)
 {
-  u_int sth = (addr.dest << 16) + addr.source;
-  u_int hash = 0;
-  
-  sth ^= addr.saddr;
-  sth += addr.daddr;
-  do {
-    // taken from scanlogd.c
-    hash ^= sth;
-  }
-  while ((sth >>= hash_deg));
-  hash %= tcp_stream_table_size;
-
-  return hash;
+  int hash=mkhash(addr.saddr, addr.source, addr.daddr, addr.dest);
+  return hash % tcp_stream_table_size;
 }
 
 static void
@@ -321,7 +322,7 @@ add_from_skb(struct tcp_stream * a_tcp, struct half_stream * rcv,
     }
   }
   else {
-    if (datalen > 0) {
+    if (datalen - lost > 0) {
       if (rcv->collect) {
 	add2buf(rcv, data + lost, datalen - lost);
 	notify(a_tcp, rcv, snd);
@@ -528,12 +529,6 @@ process_tcp(u_char * data, int skblen)
     return;
   } // ktos sie bawi
 
-  if (my_tcp_check(this_tcphdr, iplen - 4 * this_iphdr->ip_hl,
-		   this_iphdr->ip_src.s_addr, this_iphdr->ip_dst.s_addr)) {
-    nids_params.syslog(NIDS_WARN_TCP, NIDS_WARN_TCP_HDR, this_iphdr,
-		       this_tcphdr);
-    return;
-  }
   if ((this_iphdr->ip_src.s_addr | this_iphdr->ip_dst.s_addr) == 0) {
     nids_params.syslog(NIDS_WARN_TCP, NIDS_WARN_TCP_HDR, this_iphdr,
 		       this_tcphdr);
@@ -541,6 +536,14 @@ process_tcp(u_char * data, int skblen)
   }
   if (!(this_tcphdr->th_flags & TH_ACK))
     detect_scan(this_iphdr);
+  if (!nids_params.n_tcp_streams) return;
+  if (my_tcp_check(this_tcphdr, iplen - 4 * this_iphdr->ip_hl,
+		   this_iphdr->ip_src.s_addr, this_iphdr->ip_dst.s_addr)) {
+    nids_params.syslog(NIDS_WARN_TCP, NIDS_WARN_TCP_HDR, this_iphdr,
+		       this_tcphdr);
+    return;
+  }
+
   check_flags(this_iphdr, this_tcphdr);
   if (!(a_tcp = find_stream(this_tcphdr, this_iphdr, &from_client))) {
     if ((this_tcphdr->th_flags & TH_SYN) &&
@@ -679,20 +682,13 @@ nids_register_tcp(void (*x))
 int
 tcp_init(int size)
 {
-  int tmp, i;
-
+  int i;
+  if (!nids_params.n_tcp_streams) return 0;
   tcp_stream_table_size = nids_params.n_tcp_streams;
   tcp_stream_table = malloc(tcp_stream_table_size * sizeof(char *));
   if (!tcp_stream_table)
     nids_params.no_mem("tcp_init");
   memset(tcp_stream_table, 0, tcp_stream_table_size * sizeof(char *));
-  tmp = tcp_stream_table_size;
-  hash_deg = 32;
-  // computing ceiling of binary logarithm of tcp_stream_table_size
-  while (!(0x80000000 & tmp)) {
-    hash_deg--;
-    tmp <<= 1;
-  }
   max_stream = 3 * tcp_stream_table_size / 4;
   streams_pool = (struct tcp_stream *) malloc((max_stream + 1) * sizeof(struct tcp_stream));
   if (!streams_pool)
@@ -701,16 +697,34 @@ tcp_init(int size)
     streams_pool[i].next_free = &(streams_pool[i + 1]);
   streams_pool[max_stream].next_free = 0;
   free_streams = streams_pool;
-
+  init_hash(); 
   return 0;
 }
 
+#if HAVE_ICMPHDR
+#define STRUCT_ICMP struct icmphdr
+#define ICMP_CODE   code
+#define ICMP_TYPE   type
+#else
+#define STRUCT_ICMP struct icmp
+#define ICMP_CODE   icmp_code
+#define ICMP_TYPE   icmp_type
+#endif
+
+#ifndef ICMP_DEST_UNREACH
+#define ICMP_DEST_UNREACH ICMP_UNREACH
+#define ICMP_PROT_UNREACH ICMP_UNREACH_PROTOCOL
+#define ICMP_PORT_UNREACH ICMP_UNREACH_PORT
+#define NR_ICMP_UNREACH   ICMP_MAXTYPE
+#endif				
+
+
 void
 process_icmp(u_char * data)
 {
   struct ip *iph = (struct ip *) data;
   struct ip *orig_ip;
-  struct icmp *pkt;
+  STRUCT_ICMP *pkt;
   struct tcphdr *th;
   struct half_stream *hlf;
   int match_addr;
@@ -720,12 +734,12 @@ process_icmp(u_char * data)
   int from_client;
   int len = ntohs(iph->ip_len) - (iph->ip_hl << 2);
   
-  if (len < sizeof(struct icmp))
+  if (len < sizeof(STRUCT_ICMP))
     return;
-  pkt = (struct icmp *) (data + (iph->ip_hl << 2));
+  pkt = (STRUCT_ICMP *) (data + (iph->ip_hl << 2));
   if (ip_compute_csum((char *) pkt, len))
     return;
-  if (pkt->icmp_type != ICMP_DEST_UNREACH)
+  if (pkt->ICMP_TYPE != ICMP_DEST_UNREACH)
     return;
   len -= 8;
   // sizeof(struct icmp) is not what we want here
@@ -737,12 +751,12 @@ process_icmp(u_char * data)
   len -= orig_ip->ip_hl << 2;
   if (len < 8)
     return;
-  if ((pkt->icmp_code & 15) == ICMP_PROT_UNREACH ||
-      (pkt->icmp_code & 15) == ICMP_PORT_UNREACH)
+  if ((pkt->ICMP_CODE & 15) == ICMP_PROT_UNREACH ||
+      (pkt->ICMP_CODE & 15) == ICMP_PORT_UNREACH)
     match_addr = 1;
   else
     match_addr = 0;
-  if (pkt->icmp_code > NR_ICMP_UNREACH)
+  if (pkt->ICMP_CODE > NR_ICMP_UNREACH)
     return;
   if (match_addr && (iph->ip_src.s_addr != orig_ip->ip_dst.s_addr))
     return;
